# This file is part of the Blockchain Data Trading Simulator
#    https://gitlab.com/MatthiasLohr/bdtsim
#
# Copyright 2020 Matthias Lohr <mail@mlohr.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import time
from typing import Any, Callable, List, Optional
from web3 import Web3
from web3.datastructures import AttributeDict
from web3.gas_strategies.time_based import fast_gas_price_strategy
from web3.providers.base import BaseProvider
from web3.types import Wei, TxParams

from ..participant import Participant
from ..contract import Contract

logger = logging.getLogger(__name__)


class TransactionLogEntry(object):
    def __init__(self, account: Participant, transaction_receipt: AttributeDict[str, Any],
                 timestamp: Optional[float] = None):
        # TODO save contract method and contract method args
        self._account = account
        self._transaction_receipt = transaction_receipt
        self._timestamp = timestamp

    @property
    def account(self) -> Participant:
        return self._account

    @property
    def transaction_receipt(self) -> AttributeDict[str, Any]:
        return self._transaction_receipt

    @property
    def timestamp(self) -> Optional[float]:
        return self._timestamp


class Environment(object):
    def __init__(self, web3_provider: BaseProvider, chain_id: int, gas_price: Optional[int] = None,
                 gas_price_strategy: Optional[Callable[[Web3, Optional[TxParams]], Wei]] = None,
                 tx_wait_timeout: int = 120) -> None:
        self._web3 = Web3(web3_provider)

        self._chain_id = chain_id
        self._gas_price = gas_price
        if gas_price_strategy is not None:
            self._web3.eth.setGasPriceStrategy(gas_price_strategy)
        else:
            self._web3.eth.setGasPriceStrategy(fast_gas_price_strategy)
        self._tx_wait_timeout = tx_wait_timeout

        self._contract: Optional[Contract] = None
        self._contract_address: Optional[str] = None

        self._transaction_logs: List[TransactionLogEntry] = []

    @property
    def chain_id(self) -> int:
        return self._chain_id

    @property
    def gas_price(self) -> Optional[int]:
        return self._gas_price

    @property
    def tx_wait_timeout(self) -> int:
        return self._tx_wait_timeout

    @property
    def transaction_logs(self) -> List[TransactionLogEntry]:
        return self._transaction_logs

    def clear_transaction_logs(self) -> None:
        self._transaction_logs = []

    def deploy_contract(self, account: Participant, contract: Contract) -> None:
        web3_contract = self._web3.eth.contract(abi=contract.abi, bytecode=contract.bytecode)
        tx_receipt = self._send_transaction(account, web3_contract.constructor())
        self._contract_address = tx_receipt['contractAddress']
        logger.debug('New contract address: %s' % self._contract_address)
        self._contract = contract

    def send_contract_transaction(self, account: Participant, method: str, *args: Any, value: int = 0) -> Any:
        if self._contract is None:
            raise RuntimeError('No contract available!')
        web3_contract = self._web3.eth.contract(address=self._contract_address, abi=self._contract.abi)
        contract_method = getattr(web3_contract.functions, method)
        factory = contract_method(*args)
        self._send_transaction(
            account=account,
            factory=factory,
            value=value
        )
        # TODO implement contract return value

    def send_direct_transaction(self, account: Participant, to: Participant, value: int = 0) -> None:
        self._send_transaction(
            account=account,
            to=to,
            value=value
        )

    def _send_transaction(self, account: Participant, factory: Optional[Any] = None, to: Optional[Participant] = None,
                          value: int = 0) -> AttributeDict[str, Any]:
        tx_dict = {
            'from': account.wallet_address,
            'nonce': self._web3.eth.getTransactionCount(account.wallet_address, 'pending'),
            'value': value,
            'chainId': self._chain_id
        }
        if to is not None:
            tx_dict['to'] = to.wallet_address
        if factory is not None:
            tx_dict = factory.buildTransaction(tx_dict)
        else:
            tx_dict['gas'] = 21000
        if self._gas_price is not None:
            tx_dict['gasPrice'] = self._gas_price
        else:
            tx_dict['gasPrice'] = self._web3.eth.generateGasPrice(tx_dict)
        tx_signed = self._web3.eth.account.sign_transaction(tx_dict, private_key=account.wallet_private_key)
        tx_hash = self._web3.eth.sendRawTransaction(tx_signed.rawTransaction)
        tx_receipt = self._web3.eth.waitForTransactionReceipt(tx_hash, self.tx_wait_timeout)
        logger.debug('Successfully submitted transaction: %s' % str(tx_receipt))
        self._transaction_logs.append(TransactionLogEntry(account, tx_receipt, time.time()))
        return tx_receipt



# This file is part of the Blockchain Data Trading Simulator
#    https://gitlab.com/MatthiasLohr/bdtsim
#
# Copyright 2020 Matthias Lohr <mail@mlohr.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import solcx  # type: ignore
from typing import Any, Dict, Optional, Tuple

SOLC_VERSION = 'v0.6.1'

logger = logging.getLogger(__name__)


class Contract(object):
    @property
    def abi(self) -> Dict[str, Any]:
        raise NotImplementedError()

    @property
    def bytecode(self) -> str:
        raise NotImplementedError()


class SolidityContract(Contract):
    def __init__(self, contract_file: str, contract_name: str) -> None:
        self._contract_file = contract_file
        self._contract_name = contract_name
        self._abi: Optional[Dict[str, Any]] = None
        self._bytecode: Optional[str] = None

    @property
    def abi(self) -> Dict[str, Any]:
        if self._abi is None:
            self._abi, self._bytecode = self.compile(self._contract_file, self._contract_name)
        return self._abi

    @property
    def bytecode(self) -> str:
        if self._bytecode is None:
            self._abi, self._bytecode = self.compile(self._contract_file, self._contract_name)
        return self._bytecode

    @staticmethod
    def compile(contract_file: str, contract_name: str, solc_version: str = SOLC_VERSION) -> Tuple[Dict[str, Any], str]:
        # configure solc
        logger.debug('Checking for solc version %s' % solc_version)
        if solc_version not in solcx.get_installed_solc_versions():
            logger.debug('solc %s not found, installing...' % solc_version)
            solcx.install_solc(solc_version)
        solcx.set_solc_version(solc_version, silent=True)

        # read contract code
        with open(contract_file, 'r') as f:
            contract_code = f.read()

        compile_result = solcx.compile_source(contract_code)['<stdin>:' + contract_name]
        return compile_result.get('abi'), compile_result.get('bin')



# This file is part of the Blockchain Data Trading Simulator
#    https://gitlab.com/MatthiasLohr/bdtsim
#
# Copyright 2020 Matthias Lohr <mail@mlohr.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from bdtsim.participant import seller, buyer
from bdtsim.protocol_path import ProtocolPath, Decision


class DecisionTest(unittest.TestCase):
    def test_equals(self):
        self.assertEqual(Decision(True, seller), Decision(True, seller))
        self.assertEqual(Decision(False, seller), Decision(False, seller))

        self.assertNotEqual(Decision(True, seller), Decision(True, buyer))
        self.assertNotEqual(Decision(False, seller), Decision(False, buyer))
        self.assertNotEqual(Decision(True, seller), Decision(False, seller))
        self.assertNotEqual(Decision(False, seller), Decision(True, seller))


class ProtocolPathTest(unittest.TestCase):
    def test_get_alternative_decision_list(self):
        # two rounds with empty paths
        for protocol_path in ProtocolPath(), ProtocolPath([]):
            self.assertEqual(None, protocol_path.get_alternative_decision_list())

            result = protocol_path.decide(seller)
            self.assertEqual([Decision(not result, seller)], protocol_path.get_alternative_decision_list())

            protocol_path.decide(buyer)
            self.assertEqual([Decision(not result, seller)], protocol_path.get_alternative_decision_list())

        # new round with pre-defined path
        protocol_path = ProtocolPath([Decision(True, seller), Decision(False, buyer)])
        protocol_path.decide(seller)

        # in the predefined list, the 2nd decision is from buyer, so seller can't decide here
        self.assertRaises(ValueError, protocol_path.decide, seller)

        protocol_path.decide(buyer)

        # Now comes a new decision
        result = protocol_path.decide(seller)
        self.assertEqual([
            Decision(True, seller),
            Decision(False, buyer),
            Decision(not result, seller)
        ], protocol_path.get_alternative_decision_list())

        protocol_path.decide(buyer)
        self.assertEqual([
            Decision(True, seller),
            Decision(False, buyer),
            Decision(not result, seller)
        ], protocol_path.get_alternative_decision_list())



# This file is part of the Blockchain Data Trading Simulator
#    https://gitlab.com/MatthiasLohr/bdtsim
#
# Copyright 2020 Matthias Lohr <mail@mlohr.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, Type
from .protocol import Protocol


class ProtocolRegistration(object):
    def __init__(self, cls: Type[Protocol], *args: Any, **kwargs: Any) -> None:
        self.cls = cls
        self.args = args
        self.kwargs = kwargs


class ProtocolManager(object):
    protocols: Dict[str, ProtocolRegistration] = {}

    def __init__(self) -> None:
        raise NotImplementedError('This class is not to be instantiated')

    @staticmethod
    def register(name: str, cls: Type[Protocol], *args: Any, **kwargs: Any) -> None:
        if not issubclass(cls, Protocol):
            raise ValueError('Provided class is not a subclass of Protocol')
        ProtocolManager.protocols[name] = ProtocolRegistration(cls, *args, **kwargs)

    @staticmethod
    def instantiate(name: str, **kwargs: Any) -> Protocol:
        environment = ProtocolManager.protocols[name]
        return environment.cls(*environment.args, **{**environment.kwargs, **kwargs})



# This file is part of the Blockchain Data Trading Simulator
#    https://gitlab.com/MatthiasLohr/bdtsim
#
# Copyright 2020 Matthias Lohr <mail@mlohr.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import json
import yaml
from typing import Dict
from .command_manager import SubCommand
from ..data_provider import DataProviderManager
from ..environment import EnvironmentManager
from ..participant import seller, buyer
from ..protocol import ProtocolManager
from ..result import ResultSerializer
from ..simulation import Simulation


class RunSubCommand(SubCommand):
    def __init__(self, parser: argparse.ArgumentParser) -> None:
        super(RunSubCommand, self).__init__(parser)
        parser.add_argument('protocol', choices=ProtocolManager.protocols.keys())
        parser.add_argument('environment', choices=EnvironmentManager.environments.keys())
        # parser.add_argument('--data-provider', choices=DataProviderManager.data_providers.keys(),
        #                     default='GenericDataProvider')
        parser.add_argument('-c', '--chain-id', type=int, default=1)
        parser.add_argument('-o', '--output-format', choices=['human-readable', 'json', 'yaml'],
                            default='human-readable')
        parser.add_argument('--gas-price', type=int, default=None)
        parser.add_argument('--tx-wait-timeout', type=int, default=120)
        parser.add_argument('-p', '--protocol-parameter', nargs=2, action='append', dest='protocol_parameters',
                            default=[])
        parser.add_argument('-e', '--environment-parameter', nargs=2, action='append', dest='environment_parameters',
                            default=[])
        # parser.add_argument('-d', '--data-provider-parameter', nargs=2, action='append',
        #                     dest='data_provider_parameters', default=[])

    def __call__(self, args: argparse.Namespace) -> None:
        protocol_parameters: Dict[str, str] = {}
        for key, value in args.protocol_parameters:
            protocol_parameters[key.replace('-', '_')] = value

        environment_parameters: Dict[str, str] = {}
        for key, value in args.environment_parameters:
            environment_parameters[key.replace('-', '_')] = value

        # data_provider_parameters: Dict[str, str] = {}
        # for key, value in args.data_provider_parameters:
        #    data_provider_parameters[key.replace('-', '_')] = value

        protocol = ProtocolManager.instantiate(
            args.protocol,
            **protocol_parameters
        )

        environment = EnvironmentManager.instantiate(
            args.environment,
            chain_id=args.chain_id,
            gas_price=args.gas_price,
            tx_wait_timeout=args.tx_wait_timeout,
            **environment_parameters
        )

        # TODO implement support for data providers
        # data_provider = DataProviderManager.instantiate(
        #     args.data_provider,
        #     **data_provider_parameters
        # )

        simulation = Simulation(
            protocol=protocol,
            environment=environment,
            data_provider=DataProviderManager.instantiate('GenericDataProvider')  # TODO replace with data_provider
        )

        simulation_result = simulation.run()
        if args.output_format == 'human-readable':
            preparation_result = simulation_result.preparation_result
            if preparation_result is not None:
                print('Preparation Costs (gas): %d (%d transaction(s))' % (
                    preparation_result.transaction_cost_sum,
                    preparation_result.transaction_count
                ))
            honest_iteration = simulation_result.completely_honest_iteration_result
            for participant in seller, buyer:
                print('%s transaction fees: %d (%d transaction(s))' % (
                    participant.name,
                    honest_iteration.transaction_fee_sum_by_participant(participant),
                    honest_iteration.transaction_count_by_participant(participant)
                ))
                # TODO calculate and show risk
        elif args.output_format == 'json':
            print(json.dumps(simulation_result, cls=ResultSerializer, indent=2))
        elif args.output_format == 'yaml':
            json_str = json.dumps(simulation_result, cls=ResultSerializer, indent=2)
            json_structure = json.loads(json_str)
            print(yaml.dump(json_structure))



# This file is part of the Blockchain Data Trading Simulator
#    https://gitlab.com/MatthiasLohr/bdtsim
#
# Copyright 2020 Matthias Lohr <mail@mlohr.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import time
from eth_tester import EthereumTester, PyEVMBackend  # type: ignore
from eth_tester.backends import pyevm  # type: ignore
from typing import Callable, Optional
from web3 import EthereumTesterProvider, Web3
from web3.types import TxParams, Wei

from . import Environment, EnvironmentManager
from ..participant import operator, seller, buyer

logger = logging.getLogger(__name__)


class PyEVMEnvironment(Environment):
    def __init__(self, chain_id: int, gas_price: Optional[int] = None,
                 gas_price_strategy: Optional[Callable[[Web3, Optional[TxParams]], Wei]] = None,
                 tx_wait_timeout: int = 120) -> None:
        if chain_id != 61:
            logger.warning('Ignoring chainId %d since PyEVM always uses chainId 61' % chain_id)

        self._pyevm_instance = self.create_pyevm_instance()
        self._eth_tester_instance = self.create_eth_tester_instance(self._pyevm_instance)

        if gas_price_strategy is None:
            gas_price_strategy = pyevm_gas_price_strategy

        super(PyEVMEnvironment, self).__init__(
            EthereumTesterProvider(self._eth_tester_instance),
            chain_id=61,
            gas_price=gas_price,
            gas_price_strategy=gas_price_strategy,
            tx_wait_timeout=tx_wait_timeout
        )

        for account_index, recipient in [(0, operator), (1, seller), (2, buyer)]:
            account = self._eth_tester_instance.get_accounts()[account_index]
            self._eth_tester_instance.send_transaction({
                'from': account,
                'to': recipient.wallet_address,
                'gas': 21000,
                'value': self._eth_tester_instance.get_balance(account)-21000
            })

    @staticmethod
    def create_eth_tester_instance(pyevm_instance: PyEVMBackend) -> EthereumTester:
        return EthereumTester(pyevm_instance)

    @staticmethod
    def create_pyevm_instance() -> PyEVMBackend:
        return PyEVMBackend({
            'bloom': 0,
            'coinbase': pyevm.main.GENESIS_COINBASE,
            'difficulty': pyevm.main.GENESIS_DIFFICULTY,
            'extra_data': pyevm.main.GENESIS_EXTRA_DATA,
            'gas_limit': 8000000,
            'gas_used': 0,
            'mix_hash': pyevm.main.GENESIS_MIX_HASH,
            'nonce': pyevm.main.GENESIS_NONCE,
            'block_number': pyevm.main.GENESIS_BLOCK_NUMBER,
            'parent_hash': pyevm.main.GENESIS_PARENT_HASH,
            'receipt_root': pyevm.main.BLANK_ROOT_HASH,
            'timestamp': int(time.time()),
            'transaction_root': pyevm.main.BLANK_ROOT_HASH,
            'uncles_hash': pyevm.main.EMPTY_RLP_LIST_HASH
        })


def pyevm_gas_price_strategy(web3: Web3, transaction_params: Optional[TxParams]) -> Wei:
    """
    https://web3py.readthedocs.io/en/stable/gas_price.html#creating-a-gas-price-strategy
    """
    return Wei(1000000000)


EnvironmentManager.register('PyEVM', PyEVMEnvironment)



# This file is part of the Blockchain Data Trading Simulator
#    https://gitlab.com/MatthiasLohr/bdtsim
#
# Copyright 2020 Matthias Lohr <mail@mlohr.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from typing import Optional
from ..contract import Contract
from ..environment import Environment
from ..participant import Participant
from ..protocol_path import ProtocolPath


class Protocol(object):
    def __init__(self, contract_is_reusable: bool = False) -> None:
        self._contract_is_reusable = contract_is_reusable

    def prepare_environment(self, environment: Environment, operator: Participant) -> None:
        pass

    def run(self, protocol_path: ProtocolPath, environment: Environment, seller: Participant, buyer: Participant)\
            -> None:
        raise NotImplementedError()

    def cleanup_environment(self, environment: Environment, operator: Participant) -> None:
        pass

    @property
    def contract_is_reusable(self) -> bool:
        return self._contract_is_reusable

    @property
    def contract(self) -> Optional[Contract]:
        raise NotImplementedError()

    @staticmethod
    def contract_path(file_var: str, filename: str) -> str:
        return os.path.join(os.path.dirname(file_var), filename)



# This file is part of the Blockchain Data Trading Simulator
#    https://gitlab.com/MatthiasLohr/bdtsim
#
# Copyright 2020 Matthias Lohr <mail@mlohr.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import Optional
from .. import Protocol, ProtocolManager
from ...contract import SolidityContract
from ... import Environment, Participant
from ...protocol_path import ProtocolPath

logger = logging.getLogger(__name__)


class SimplePayment(Protocol):
    def __init__(self, use_contract: bool) -> None:
        super(SimplePayment, self).__init__(contract_is_reusable=True)

        self._use_contract = use_contract

    @property
    def contract(self) -> Optional[SolidityContract]:
        if self._use_contract:
            return SolidityContract(self.contract_path(__file__, 'SimplePayment.sol'), 'SimplePayment')
        else:
            return None

    def run(self, protocol_path: ProtocolPath, environment: Environment, seller: Participant, buyer: Participant)\
            -> None:
        if protocol_path.decide(buyer):
            logger.debug('Decided to be honest')
            if self._use_contract:                                                       # TODO parameterize value
                environment.send_contract_transaction(buyer, 'pay', seller.wallet_address, value=1000000000)
            else:
                environment.send_direct_transaction(buyer, seller, 1000000000)  # TODO parameterize value
        else:
            logger.debug('Decided to cheat')  # do nothing


ProtocolManager.register('SimplePayment', SimplePayment, use_contract=True)
ProtocolManager.register('SimplePayment-direct', SimplePayment, use_contract=False)



# This file is part of the Blockchain Data Trading Simulator
#    https://gitlab.com/MatthiasLohr/bdtsim
#
# Copyright 2020 Matthias Lohr <mail@mlohr.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from hexbytes.main import HexBytes
from json import JSONEncoder
from typing import Any, Dict, List, Optional
from web3.datastructures import AttributeDict
from .environment.environment import TransactionLogEntry
from .participant import Participant
from .protocol_path import ProtocolPath, Decision


class PreparationResult(object):
    def __init__(self, transaction_logs: Optional[List[TransactionLogEntry]] = None) -> None:
        self._transaction_logs = transaction_logs or []

    @property
    def transaction_logs(self) -> List[TransactionLogEntry]:
        return self._transaction_logs

    @property
    def transaction_count(self) -> int:
        return len(self.transaction_logs)

    @property
    def transaction_cost_sum(self) -> int:
        return sum([int(str(entry.transaction_receipt.get('gasUsed'))) for entry in self.transaction_logs])


class IterationResult(object):
    def __init__(self, protocol_path: ProtocolPath, transaction_logs: Optional[List[TransactionLogEntry]] = None)\
            -> None:
        self._protocol_path = protocol_path
        self._transaction_logs = transaction_logs or []

    @property
    def protocol_path(self) -> ProtocolPath:
        return self._protocol_path

    @property
    def transaction_logs(self) -> List[TransactionLogEntry]:
        return self._transaction_logs

    @property
    def is_completely_honest(self) -> bool:
        return self.protocol_path.is_completely_honest

    def transaction_fee_sum_by_participant(self, participant: Participant) -> int:
        return sum([
            int(str(log.transaction_receipt.get('gasUsed')))
            for log in filter(lambda x: x.account == participant, self._transaction_logs)
        ])

    def transaction_count_by_participant(self, participant: Participant) -> int:
        return len(list(filter(lambda x: x.account == participant, self._transaction_logs)))


class SimulationResult(object):
    def __init__(self, operator: Participant, seller: Participant, buyer: Participant,
                 preparation_result: Optional[PreparationResult] = None,
                 iteration_results: Optional[List[IterationResult]] = None) -> None:
        self._participants_by_address: Dict[str, Participant] = {}
        for participant in operator, seller, buyer:
            self._participants_by_address.update({participant.wallet_address: participant})

        self._preparation_result = preparation_result
        self._iteration_results = iteration_results or []

    @property
    def preparation_result(self) -> Optional[PreparationResult]:
        return self._preparation_result

    @property
    def iteration_results(self) -> List[IterationResult]:
        return self._iteration_results

    @property
    def completely_honest_iteration_result(self) -> IterationResult:
        for ir in self.iteration_results:
            if ir.is_completely_honest:
                return ir
        raise RuntimeError('Impossible result: There MUST be a completely honest iteration')


class ResultSerializer(JSONEncoder):
    def default(self, obj: Any) -> Any:
        if isinstance(obj, AttributeDict):
            return dict(obj)
        elif isinstance(obj, Decision):
            return {
                'account': obj.account,
                'decision': obj.decision,
                'timestamp': obj.timestamp
            }
        elif isinstance(obj, HexBytes):
            return obj.hex()
        elif isinstance(obj, IterationResult):
            return {
                'protocol_path': obj.protocol_path,
                'transaction_logs': obj.transaction_logs
            }
        elif isinstance(obj, Participant):
            return {
                'name': obj.name,
                'wallet_address': obj.wallet_address
            }
        elif isinstance(obj, PreparationResult):
            return {
                'transaction_logs': obj.transaction_logs
            }
        elif isinstance(obj, ProtocolPath):
            return obj.decisions_list
        elif isinstance(obj, SimulationResult):
            return {
                'preparation_result': obj.preparation_result,
                'iteration_results': obj.iteration_results
            }
        elif isinstance(obj, TransactionLogEntry):
            return {
                'account': obj.account,
                'transaction_receipt': obj.transaction_receipt,
                'timestamp': obj.timestamp
            }
        else:
            return JSONEncoder.default(self, obj)



# This file is part of the Blockchain Data Trading Simulator
#    https://gitlab.com/MatthiasLohr/bdtsim
#
# Copyright 2020 Matthias Lohr <mail@mlohr.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from queue import Queue
from .participant import operator, seller, buyer
from .data_provider import DataProvider
from .environment import Environment
from .protocol import Protocol
from .protocol_path import ProtocolPath
from .result import PreparationResult, IterationResult, SimulationResult

logger = logging.getLogger(__name__)


class Simulation(object):
    def __init__(self, protocol: Protocol, environment: Environment, data_provider: DataProvider):
        self._protocol = protocol
        self._environment = environment
        self._data_provider = data_provider

        self._protocol_path_queue: "Queue[ProtocolPath]" = Queue()

    def run(self) -> SimulationResult:
        # initial contract deployment
        preparation_result = None
        self._environment.clear_transaction_logs()
        if self._protocol.contract_is_reusable and self._protocol.contract is not None:
            logger.debug('Initial contract deployment...')
            self._environment.clear_transaction_logs()
            self._environment.deploy_contract(operator, self._protocol.contract)
            preparation_result = PreparationResult(self._environment.transaction_logs)
            logger.debug('Initial contract deployment finished')

        iteration_results = []
        self._protocol_path_queue.put(ProtocolPath())
        logger.debug('Simulation started')
        while not self._protocol_path_queue.empty():
            protocol_path = self._protocol_path_queue.get(block=False)
            self._environment.clear_transaction_logs()
            logger.debug('Simulation will follow path %s' % str(protocol_path.decisions_list))

            # per-iteration contract deployment
            if not self._protocol.contract_is_reusable and self._protocol.contract is not None:
                logger.debug('Iteration contract deployment...')
                self._environment.deploy_contract(operator, self._protocol.contract)
                logger.debug('Iteration contract deployment finished')

            logger.debug('Preparing environment...')
            self._protocol.prepare_environment(self._environment, operator)
            logger.debug('Starting protocol iteration...')
            self._protocol.run(protocol_path, self._environment, seller, buyer)
            logger.debug('Protocol iteration finished')
            iteration_results.append(IterationResult(
                protocol_path=protocol_path,
                transaction_logs=self._environment.transaction_logs
            ))
            logger.debug('Collecting alternative paths...')
            alternative_decision_list = protocol_path.get_alternative_decision_list()
            if alternative_decision_list is not None:
                self._protocol_path_queue.put(ProtocolPath(alternative_decision_list))
                logger.debug('Added new path %s' % str(alternative_decision_list))
            logger.debug('Cleaning up environment...')
            self._protocol.cleanup_environment(self._environment, operator)
            self._protocol_path_queue.task_done()
        logger.debug('Simulation finished')
        return SimulationResult(
            operator=operator,
            seller=seller,
            buyer=buyer,
            preparation_result=preparation_result,
            iteration_results=iteration_results
        )



# This file is part of the Blockchain Data Trading Simulator
#    https://gitlab.com/MatthiasLohr/bdtsim
#
# Copyright 2020 Matthias Lohr <mail@mlohr.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Callable, Optional, Type
from web3 import HTTPProvider, WebsocketProvider, IPCProvider, Web3
from web3.providers.base import JSONBaseProvider
from web3.types import TxParams, Wei

from . import Environment, EnvironmentManager


class Web3Environment(Environment):
    def __init__(self, web3_provider_class: Type[JSONBaseProvider], chain_id: int, gas_price: Optional[int] = None,
                 gas_price_strategy: Optional[Callable[[Web3, Optional[TxParams]], Wei]] = None,
                 tx_wait_timeout: int = 120, **kwargs: Any) -> None:
        # noinspection PyArgumentList
        super(Web3Environment, self).__init__(
            web3_provider_class(**kwargs),  # type: ignore
            chain_id=chain_id,
            gas_price=gas_price,
            gas_price_strategy=gas_price_strategy,
            tx_wait_timeout=tx_wait_timeout
        )

    def set_up(self) -> None:
        pass

    def strip_down(self) -> None:
        pass


EnvironmentManager.register('Web3HTTP', Web3Environment, HTTPProvider)
EnvironmentManager.register('Web3Websocket', Web3Environment, WebsocketProvider)
EnvironmentManager.register('Web3IPC', Web3Environment, IPCProvider)



# This file is part of the Blockchain Data Trading Simulator
#    https://gitlab.com/MatthiasLohr/bdtsim
#
# Copyright 2020 Matthias Lohr <mail@mlohr.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
from typing import Any, List, Optional
from .participant import Participant


class Decision(object):
    def __init__(self, decision: bool, account: Participant, timestamp: Optional[float] = None) -> None:
        self._decision = decision
        self._account = account
        self._timestamp = timestamp

    @property
    def decision(self) -> bool:
        return self._decision

    @property
    def account(self) -> Participant:
        return self._account

    @property
    def timestamp(self) -> Optional[float]:
        return self._timestamp

    @timestamp.setter
    def timestamp(self, timestamp: float) -> None:
        self._timestamp = timestamp

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, Decision) and other.decision == self.decision and other.account == self.account

    def __ne__(self, other: Any) -> bool:
        return not self.__eq__(other)

    def __repr__(self) -> str:
        return '<%s.%s %s, account: %s>' % (__name__, self.__class__.__name__, str(self.decision), repr(self.account))


class ProtocolPath(object):
    def __init__(self, decisions_list: Optional[List[Decision]] = None):
        self._initial_decisions_list: List[Decision] = decisions_list or []
        self._new_decisions_list: List[Decision] = []
        self._decisions_index: int = 0

    def decide(self, account: Participant) -> bool:
        if len(self.decisions_list) == self._decisions_index:
            self._new_decisions_list.append(Decision(True, account, None))
        decision = self.decisions_list[self._decisions_index]
        decision.timestamp = time.time()
        if decision.account != account:
            raise ValueError('Predefined decision is not originating from current decider')
        self._decisions_index += 1
        return decision.decision

    @property
    def initial_decisions_list(self) -> List[Decision]:
        return self._initial_decisions_list

    @property
    def new_decisions_list(self) -> List[Decision]:
        return self._new_decisions_list

    @property
    def decisions_list(self) -> List[Decision]:
        return self.initial_decisions_list + self.new_decisions_list

    def get_alternative_decision_list(self) -> Optional[List[Decision]]:
        if len(self._new_decisions_list) > 0:
            return self.initial_decisions_list + [Decision(
                decision=not self._new_decisions_list[0].decision,
                account=self._new_decisions_list[0].account
            )]
        else:
            return None

    @property
    def is_completely_honest(self) -> bool:
        for d in self.decisions_list:
            if not d.decision:
                return False
        return True
